% Kyrre Sjobak, Sept. 2021
% For newest version and example for how to use, please see
%   https://gitlab.cern.ch/CLEAR/arduino4d_robot/
%   Subfolder 'MatLab'


classdef robotConnector < handle
    %robotConnector Class for connecting to the CLEAR 4D robot
    %   Please read the documentation-comment in the
    %   robotConnector() constructor function.
    %
    %   The status of the robot is indicated with the variables
    %   in the properties block below, e.g. stepperPos, stepperAbs, etc.
    %
    %   testClass.m contains a few very simple examples on how to use this class.

    % Note: All strings are of "new" type, not 'old'.
    %       This seems to be required by MATLAB TCP/IP library.
    
    properties (SetAccess = protected)
        %Interface status
        
        isReady = 0       %True when this interface is able to recieve commands
                          % Note: The robot may be busy and not responding
                          % immediately, as it is queuing internally.
                    
        lastCommand = ""; %The last command that the robot recieved
                          % (on any interface) while we were connected
        

        %Callback function to GUI etc.
        % Note: This print-only function would normally be replaced by the user.
        % The first argument is a pointer to the object that has been updated.
        % The three last arguments indicate which part has been updated.
        callbackFunction = @(obj,stepper,grabber,temp) ...
            disp(['UPDATE: stepper=',num2str(stepper), ...
                        ', grabber=',num2str(grabber), ...
                        ', temp=',   num2str(temp) ]);

        %Optional extra callback function which is notified whenever a line
        % of text is recieved from the robot.
        textCallbackFunction = @(obj,textLine) disp(textLine);
        
        %% Variables representing the current status of the robot
        
        %Stepper status [X/Y/Z]
        stepperXYZ       = ["X","Y","Z"]; %Names of the axis in this class
        stepperPos       = [NaN,NaN,NaN]; %Current position
        stepperAbs       = [NaN,NaN,NaN]; %Absolute mode?
        stepperLimLo     = [NaN,NaN,NaN]; %Limit switches active?
        stepperLimHi     = [NaN,NaN,NaN]; %Limit switches active?
        stepperInterlock = NaN;           %Safety interlock enabled?

        stepperIsMoving = 0;              %Which axis (1/2/3) is moving? 0 if not moving.

        %Grabber status
        grabberPos       = NaN;

        grabberClosedPos = NaN;
        grabberOpenPos   = NaN;
        grabberMinPos    = NaN;
        grabberMaxPos    = NaN;

        grabberIsMoving  = false;
        grabberGotoTarget = NaN;
        
        %Temperatures [degC]
        temps = [];

    end
    
    properties (SetAccess = immutable)
        address %Hostname or IP address of Arduino, as a string

        TCPport %TCP port of the arduino, as an int
    end
    
    properties (Access = private)
        TCPconn %TCP connection object
                % Connection is automatically destroyed when
                % robotConnector object is destroyed

        %Special variables for switching between serial (blocking) mode
        % and callback mode
        waitingForReady = false;  %Used by waitForReady()

        %If this is true, callback from TCP to parser is disabled, so
        %parser must be manually triggered. Furthermore the parser will not
        %call obj.callbackFunction.
        callbackEnabled = false;

        %Used by parser:
        linesWithLastCommand = 0; %The number of output (<) lines we have gotten
                                  % since lastCommand
                                  % (the command echo is line #0)
                                  
        commandIsRunning = false; %Gotten all the expected output from a command?
    end
    
    %% USER METHODS

    methods
        function obj = robotConnector(address, callbackFunction, textCallbackFunction, TCPport)
            %robotConnector: Construct and return an instance of this class,
            %                and open a connection to the robot.
            %
            %   Mandatory arguments:
            %   - address :             The hostname or IP address to connect to, as a string.
            %                             E.g. '192.168.0.42', 'localhost', '4drobot.domain.com', ...
            %
            %   Optional arguments:
            %
            %   - callbackFunction:     Handle to afunction that is called whenever
            %                             there is an update of the object status.
            %                             Function signature: @(obj,stepper,grabber,temp)
            %                             The called function can then read the
            %                             status from obj.stepperPos etc.
            %                             Note that if using the _blocking
            %                             functions, this callback is NOT called.
            %
            %   - textCallbackFunction: Handle to a function that is called whenever
            %                             a line of text is recieved from the robot.
            %                             Function signature: @(obj,textLine)
            %                             The called function gets the last recieved line of text
            %                             as the argument `textLine`;
            %                             by default it is just displayed with `disp()`.
            %
            %   - port:                 The TCP port on address to connect to.
            %                             Defaults to 23 if not provided or set to NaN.
            %                             This is useful for debugging, e.g. redirecting the
            %                             class to a local listener such as `netcat -l 4223`.
            %
            
            %Setup callback
            if exist('callbackFunction','var')
                if ~isa(callbackFunction,'function_handle')
                    error('Argument callbackFunction must be a function handle taking 4 arguments, obj and 3 bools')
                end
                obj.callbackFunction = callbackFunction;
            end

            %Setup textline callback
            if exist('textCallbackFunction','var')
                if ~isa(textCallbackFunction,'function_handle')
                    error('Argument textCallbackFunction must be a function handle taking 2 arguments, obj and textline.')
                end
                obj.textCallbackFunction = textCallbackFunction;
            end

            %Setup connection
            obj.address = address;
            if exist('TCPport','var')
                obj.TCPport = TCPport;
            else
                obj.TCPport = 23;
            end
            
            disp(['Connecting to ', obj.address, ' on port ', num2str(obj.TCPport), '...']);
            obj.TCPconn = tcpclient(obj.address,obj.TCPport, "Timeout", 2);
            disp('Connected!')

            obj.TCPconn.configureTerminator('LF')

            %Wait untill a '$' has been seen before
            %starting to issue commands...
            obj.waitForReady();
            
            %Ready to go!
            obj.enableCallbacks();
            
            %Request population of data
            obj.updateStepperStatus();
            obj.updateGrabberStatus();
            obj.updateTemp();
        end
        
        function delete(obj)
            %DELETE Closes the TCP connection
            disp(['Disconnecting from ', obj.address, ' on port ', num2str(obj.TCPport), '...']);
            clear obj.TCPconn;
        end
        
        %% Commands that can be queued up

        % Trigger download of data
        % Note: The robot may take some time to respond,
        % so please wait for a callback before reading the data!

        function [ok] = updateStepperStatus(obj)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            
            obj.TCPconn.write(['STEPPER STATUS',newline]);
            
            ok=true;
            return
        end

        function [ok] = updateGrabberStatus(obj)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            
            obj.TCPconn.write(['GRABBER STATUS',newline]);
            
            ok=true;
            return
        end

        function [ok] = updateTemp(obj)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            
            obj.TCPconn.write(['TEMP',newline]);
            
            ok=true;
            return
        end

        % Send an active command
        % Data is returned by callback

        function [ok] = grabberGo(obj,pos)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            
            if ~exist('pos','var')
                error('Input POS must be set')
            end
            if ~(isnumeric(pos) && length(pos) == 1)
                error('Input POS must be a single integer number')
            end

            if ~(pos >= obj.grabberMinPos && pos <= obj.grabberMaxPos)
                error('Requested grabber pos out of range')
            end
            obj.TCPconn.write(['GRABBER POS ', num2str(round(pos)), newline]);

            ok=true;
            return
        end

        function [ok] = grabberOpen(obj)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end

            obj.TCPconn.write(['GRABBER OPEN',newline]);

            ok=true;
            return
        end

        function [ok] = grabberClose(obj)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end

            obj.TCPconn.write(['GRABBER CLOSE',newline]);

            ok=true;
            return
        end
        
        function [ok] = stepperZeroseek(obj, xyz)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end

            if isempty(find(obj.stepperXYZ == xyz, 1))
                error(strcat("xyz must be in ", join(obj.stepperXYZ), " got ",xyz))
            end

            obj.TCPconn.write(strcat("STEPPER ZEROSEEK ", string(xyz), string(newline)));

            ok=true;
            return
        end

        function [ok] = stepperGo(obj, xyz, pos)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            
            if isempty(find(obj.stepperXYZ == xyz, 1))
                error(strcat("xyz must be in ", join(obj.stepperXYZ), " got ",xyz))
            end
            
            if ~exist('pos','var')
                error('Input POS must be set')
            end
            if ~(isnumeric(pos) && length(pos) == 1)
                error('Input POS must be a single integer number')
            end

            obj.TCPconn.write(strcat("STEPPER POS ", string(xyz), " ", string(num2str(round(pos))), string(newline)));

            ok=true;
            return
        end

        % Send a config change command
        % Data is returned by callback

        function [ok] = setStepperInterlock(obj,makeSafe)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end

            obj.TCPconn.write('STEPPER INTERLOCK ');
            if makeSafe
                obj.TCPconn.write(['ENABLE', newline]);
            else
                obj.TCPconn.write(['DISABLE', newline]);
            end

            ok=true;
            return
        end

        function [ok] = setStepperReset(obj,xyz)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end

            if isempty(find(obj.stepperXYZ == xyz, 1))
                error(strcat("xyz must be in ", join(obj.stepperXYZ), " got ",xyz))
            end

            obj.TCPconn.write(strcat("STEPPER RESET ", string(xyz), string(newline)));

            %Note:
            % It may be better if this MATLAB command had come come with
            % an automatic stepper status update command,
            % which is queued on the Arduino right after the stepper reset command.
            % Currently the parser will try to reflect the status of the
            % Arduino without actually reading it.

            %ok=obj.updateStepperStatus();
            ok=true;
            return
        end

        % Send a custom command
        function [ok] = sendCustomCommand(obj,cmd)
            if ~obj.isReady
                disp('Robot interface not ready')
                ok = false;
                return;
            end
            obj.TCPconn.write(strcat(string(cmd), string(newline)));
        end
        
        %% Blocking variants of commands
        %  Note: These do not play well with multiple connected users
        %  sending commands, race conditions are VERY LIKELY
        %  However being concurrent to users that are only reading the
        %  status should be OK.
        %
        %  These functions are useful for scripting movement, e.g. go
        %  there, grab, then go somewhere else, but VERY BAD for GUIs even
        %  if they may seem easier. It is intentional that MATLAB freezes
        %  when they are called, and Control-C behaviour is UNDEFINED.
        %
        %  Note that even if the callback which normaly notifies
        %  that the state of the robotConnector object has changed is disabled,
        %  the state is still modified as normal when data is recieved from
        %  the robot. It is up to the user to read e.g. stepperPos after
        %  calling these methods.
        function updateStepperStatus_blocking(obj)
            obj.disableCallbacks();
            obj.updateStepperStatus();
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function updateGrabberStatus_blocking(obj)
            obj.disableCallbacks();
            obj.updateGrabberStatus();
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function updateTemp_blocking(obj)
            obj.disableCallbacks();
            obj.updateTemp();
            obj.waitForReady();
            obj.enableCallbacks();
        end
        
        function grabberGo_blocking(obj,pos)
            obj.disableCallbacks();
            obj.grabberGo(pos);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function grabberOpen_blocking(obj)
            obj.disableCallbacks();
            obj.grabberOpen();
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function grabberClose_blocking(obj)
            obj.disableCallbacks();
            obj.grabberClose();
            obj.waitForReady();
            obj.enableCallbacks();
        end

        function stepperZeroseek_blocking(obj,xyz)
            obj.disableCallbacks();
            obj.stepperZeroseek(xyz);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function stepperGo_blocking(obj,xyz,pos)
            obj.disableCallbacks();
            obj.stepperGo(xyz,pos);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        
        function setStepperInterlock_blocking(obj,makeSafe)
            obj.disableCallbacks();
            obj.setStepperInterlock(makeSafe);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        function setStepperReset_blocking(obj,xyz)
            obj.disableCallbacks();
            obj.setStepperReset(xyz);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        
        function sendCustomCommand_blocking(obj,cmd)
            obj.disableCallbacks();
            obj.sendCustomCommand(cmd);
            obj.waitForReady();
            obj.enableCallbacks();
        end
        
    end
    
    %% INTERNAL METHODS
    
    methods (Access = protected)
        
        function enableCallbacks(obj)
            obj.TCPconn.configureCallback("byte",1, @(srcConn, info) TCPdataCallback(obj))
            obj.callbackEnabled = true;
        end
        
        function disableCallbacks(obj)
            obj.TCPconn.configureCallback("off")
            obj.callbackEnabled = false;
        end
        
        function [ok] = waitForReady(obj)
            %This will return after a new '$' has been recieved.
            % To be used for serialized scripts that ignore the callback.
            %Note: Calling this without any data being incoming will freeze
            % the script.
            
            wasCallbacksOn = obj.callbackEnabled;
            if wasCallbacksOn
                obj.disableCallbacks()
            end
            
            if obj.waitingForReady
                disp('Only one waitForReady() can be active at a given time!')
                ok = false;
                return;
            end
            obj.waitingForReady = true;
            
            %Lazy busywait for data
            % (Note: TCPdataCallback can handle incomplete lines)
            while (obj.waitingForReady)
                pause(0.1)
                n = obj.TCPconn.BytesAvailable;
                if n > 0
                    TCPdataCallback(obj);
                end
            end
            
            if wasCallbacksOn
                obj.enableCallbacks();
            end

            ok=true;
            return
        end
        
        function TCPdataCallback(obj)
            %TCPdataCallback Recieves data from the TCP connection
            %   and updates the class status
            %disp('CB')
            
            n = obj.TCPconn.BytesAvailable;
            if n == 0
                %disp('Empty data')
                return
            end
            
            line = obj.TCPconn.readline();
            if isempty(line)
                %No newline in data or empty string returned
                %disp('Data not ready')
                return
            end
            
            %Callback function when recieving text
            % by default it's just `disp(line)`
            obj.textCallbackFunction(obj,line)
            
            %Parse the input!
                        
            if startsWith(line,'$')
                obj.isReady = true;
                obj.commandIsRunning = false;
                obj.waitingForReady = false;
                %disp('READY')
                
            elseif startsWith(line,'>')
                
                obj.lastCommand = extractAfter(line, '> ');
                obj.linesWithLastCommand = 0;
                
                %Set to true if we are expecting output to be parsed
                if obj.lastCommand == "STEPPER STATUS" || ...
                   obj.lastCommand == "GRABBER STATUS" || ...
                   obj.lastCommand == "TEMP" || ...
                   obj.lastCommand == "GRABBER OPEN" || obj.lastCommand == "GRABBER CLOSE" || startsWith(obj.lastCommand,'GRABBER POS ') || ...
                   startsWith(obj.lastCommand, 'STEPPER INTERLOCK ') || ...
                   startsWith(obj.lastCommand, 'STEPPER RESET ') || ...
                   startsWith(obj.lastCommand, 'STEPPER ZEROSEEK ') || startsWith(obj.lastCommand, 'STEPPER POS ')
               
                    if obj.commandIsRunning
                        obj.commandIsRunning = false; %Reset!
                        error('Implementation error: Command already running?')
                    end

                    %Command was recognized!
                    %Now the next lines should be output from this
                    %command... Change the state of the class!
                    obj.commandIsRunning = true;
                end
                    

            elseif startsWith(line, '< ')
                obj.linesWithLastCommand = obj.linesWithLastCommand + 1;
                
                if obj.commandIsRunning
                    % Parse expected output lines!

                    if obj.lastCommand == "STEPPER STATUS"
                        if obj.linesWithLastCommand == 1
                            if contains(line,'ENABLED')
                                obj.stepperInterlock = true;
                            elseif contains(line,'DISABLED')
                                obj.stepperInterlock = false;
                            else
                                disp('ERROR: Interlock status could not be determined')
                            end
                        elseif obj.linesWithLastCommand == 3 || obj.linesWithLastCommand == 4 || obj.linesWithLastCommand == 5
                            
                            lineSplit = split(line);
                            
                            xyz = lineSplit(3);
                            xyzID = find(obj.stepperXYZ==xyz);
                            if isempty(xyzID)
                                error(strcat("Axis name ", string(xyz), " not recognized"))
                            end
                            
                            obj.stepperPos(xyzID) = str2double(lineSplit(4));
                            obj.stepperAbs(xyzID) = logical(str2double(lineSplit(5)));
                            obj.stepperLimLo(xyzID) = logical(str2double(lineSplit(6)));
                            obj.stepperLimHi(xyzID) = logical(str2double(lineSplit(7)));
                            
                            if obj.linesWithLastCommand == 5
                                %We are done!
                                obj.commandIsRunning = false;
                                if obj.callbackEnabled
                                    obj.callbackFunction(obj,true,false,false);
                                end
                            end
                        end

                    elseif obj.lastCommand == "GRABBER STATUS"
                        lineSplit = split(line);
                        if obj.linesWithLastCommand == 1
                            obj.grabberClosedPos = str2double(lineSplit(6));
                            obj.grabberOpenPos   = str2double(lineSplit(9));
                            obj.grabberMinPos = str2double(lineSplit(13));
                            obj.grabberMaxPos   = str2double(lineSplit(16));
                        elseif obj.linesWithLastCommand == 2
                            obj.grabberPos       = str2double(lineSplit(5));
                            obj.commandIsRunning = false;
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,false,true,false);
                            end
                        end

                    elseif obj.lastCommand == "TEMP"
                        lineSplit = split(line);
                        if obj.linesWithLastCommand == 1
                            numTemps = str2double(lineSplit(4));
                            obj.temps = ones(numTemps,1)*NaN;
                        else
                            obj.temps(obj.linesWithLastCommand-1) = str2double(lineSplit(4));
                        end
                        
                        if obj.linesWithLastCommand > length(obj.temps)
                            %Data for sensor N at line N+1,
                            % so stop when N+1 is reached
                            obj.commandIsRunning = false;
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,false,false,true);
                            end
                        end
                        
                    elseif obj.lastCommand == "GRABBER OPEN"  || ...
                           obj.lastCommand == "GRABBER CLOSE" || ...
                           startsWith(obj.lastCommand,'GRABBER POS ')
                        lineSplit = split(line);
                        if obj.linesWithLastCommand == 1
                            obj.grabberGotoTarget = str2double(lineSplit(4));
                            %Notify that we have started moving
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,false,true,false);
                            end
                        elseif obj.linesWithLastCommand == 2
                            obj.grabberIsMoving = true;
                        elseif obj.linesWithLastCommand == 3
                            obj.grabberGotoTarget = NaN;
                            obj.grabberIsMoving = false;
                            obj.commandIsRunning = false;
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,false,true,false);
                            end
                        else
                            error('Implementation error!');
                        end

                    elseif startsWith(obj.lastCommand, 'STEPPER INTERLOCK ')
                        lineSplit = split(line);
                        if lineSplit(5) == "ENABLED"
                            obj.stepperInterlock = true;
                        elseif lineSplit(5) == "DISABLED"
                            obj.stepperInterlock = false;
                        else
                            error('Did not understand reply?')
                        end
                        obj.commandIsRunning = false;
                        if obj.callbackEnabled
                            obj.callbackFunction(obj,true,false,false)
                        end

                    elseif startsWith(obj.lastCommand, 'STEPPER RESET ')
                        lineSplit = split(line);
                        xyz = lineSplit(5);
                        xyzID = find(obj.stepperXYZ==xyz);
                        if isempty(xyzID)
                            error(strcat("Axis name ", string(xyz), " not recognized"))
                        end

                        obj.stepperPos(xyzID) = 0;
                        obj.stepperAbs(xyzID) = false;

                        obj.commandIsRunning = false;
                        if obj.callbackEnabled
                            obj.callbackFunction(obj,true,false,false);
                        end

                    elseif startsWith(obj.lastCommand, 'STEPPER ZEROSEEK ') || ...
                           startsWith(obj.lastCommand, 'STEPPER POS ')

                        lineSplitCMD = split(obj.lastCommand);
                        xyz = lineSplitCMD(3);
                        xyzID = find(obj.stepperXYZ==xyz);
                        if isempty(xyzID)
                            error(strcat("Axis name ", string(xyz), " not recognized"))
                        end

                        lineSplit = split(line);

                        if startsWith(line, '< STEPPER GO FINISHED') % This is always the final output line
                            obj.stepperIsMoving = 0;
                            obj.stepperPos(xyzID) = str2double(lineSplit(8));

                            if startsWith(obj.lastCommand, 'STEPPER ZEROSEEK ')
                                if obj.stepperLimLo(xyzID) == false && obj.stepperLimHi(xyzID) == false
                                    %sucess!
                                    obj.stepperLimLo(xyzID) = true;
                                    obj.stepperAbs(xyzID) = true;
                                else
                                    %Failed zeroseek.
                                    % Note that if it was in absolute mode before,
                                    % it still is in absolute mode even if we failed.
                                end
                            end

                            obj.commandIsRunning = false;
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,true,false,false);
                            end

                        elseif line.startsWith("< STEPPER MOVING ")
                            obj.stepperIsMoving = xyzID;
                            if obj.callbackEnabled
                                obj.callbackFunction(obj,true,false,false);
                            end
                            %Some of these will be set to true if
                            % (1) we touch a limit switch, or
                            % (2) we do a successful zeroseek
                            obj.stepperLimLo(xyzID) = false;
                            obj.stepperLimHi(xyzID) = false;

                        elseif line.startsWith("< STEPPER ERROR LIMIT SWITCH")
                            obj.stepperLimLo(xyzID) = logical(str2double(lineSplit(8)));
                            obj.stepperLimHi(xyzID) = logical(str2double(lineSplit(11)));

                        elseif line == "< STEPPER ERROR XY INTERLOCK"
                            assert(obj.stepperInterlock, "Status reflection error, did not expect interlock to be enabled")

                        elseif line == "< STEPPER WARNING XY INTERLOCK IGNORED"
                            assert(obj.stepperInterlock, "Status reflection error, expected interlock to be enabled")

                        end

                    else
                        error('lastCommand unknown to parser even if commandIsRunning==true?')
                    end

                else
                    %Ignore unsolicited output lines
                end

            elseif startsWith(line, '@')
                %Local error / info, ignore

            else
                error('Unknown return type!')
            end

        end
    end
end

